#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <SDL.h>

#include "metaballs.h"
#include "meshgen.h"

#define WIDTH  600
#define HEIGHT 600
#define BPP    16

static metaball_t balls[NR_METABALLS];
static float ambient[] = {0.0f, 0.0f, 0.2f, 1.0f};
static float position[] = {-1.0f, 1.0f, 1.0f, 0.0f};
static float specular[] = {1.0f, 1.0f, 1.0f, 1.0f};
static float shininess = 4.0f;
static float light_val = 1.0f;

static int polygonMode = 1;

void init_balls() {

    int i;

    for (i = 0; i < NR_METABALLS; i++) {
        balls[i].radius = (rand() % 10) + 1.0;
        
        balls[i].pos.x = balls[i].radius + (rand() % (int) (SPACE_SIZE - balls[i].radius*2));
        balls[i].pos.y = balls[i].radius + (rand() % (int) (SPACE_SIZE - balls[i].radius*2));
        balls[i].pos.z = balls[i].radius + (rand() % (int) (SPACE_SIZE - balls[i].radius*2));
        
        balls[i].strength = 1.0;
        balls[i].acc.x = (rand() % 10) * 0.0001;
        balls[i].acc.y = (rand() % 10) * 0.0001;
        balls[i].acc.z = (rand() % 10) * 0.0001;
        balls[i].vel.x = (rand() % 10) * 0.0001;
        balls[i].vel.y = (rand() % 10) * 0.0001;
        balls[i].vel.z = (rand() % 10) * 0.0001;
    }
    
    /*
    
    balls[0].pos.x = 20.0;
    balls[0].pos.y = 25.0;
    balls[0].pos.z = 25.0;
    balls[0].radius = 12.5;
    balls[0].strength = 1.0;
    balls[0].acc.x = 0.0;
    balls[0].acc.y = -0.0001;
    balls[0].acc.z = 0.0;
    balls[0].vel.x = 0.0005;
    balls[0].vel.y = 0.0;
    balls[0].vel.z = 0.0005;

    balls[1].pos.x = 10.0;
    balls[1].pos.y = 10.0;
    balls[1].pos.z = 10.0;
    balls[1].radius = 3.0;
    balls[1].strength = 1.0;
    balls[1].acc.x = 0.0;
    balls[1].acc.y = 0.0002;
    balls[1].acc.z = 0.0;
    balls[1].vel.x = -0.001;
    balls[1].vel.y = 0.0;
    balls[1].vel.z = 0.0;

    balls[2].pos.x = 15.0;
    balls[2].pos.y = 15.0;
    balls[2].pos.z = 15.0;
    balls[2].radius = 5.5;
    balls[2].strength = 1.0;
    balls[2].acc.x = -0.0003;
    balls[2].acc.y = 0.0;
    balls[2].acc.z = 0.0;
    balls[2].vel.x = 0.0;
    balls[2].vel.y = -0.001;
    balls[2].vel.z = 0.0;
    */
}


void keyHandler(SDL_keysym *key) {

    switch (key->sym) {
        case SDLK_ESCAPE:
            exit(EXIT_SUCCESS);
            break;
        case SDLK_p:
            if (polygonMode) {
                glPolygonMode(GL_FRONT, GL_LINE);
                polygonMode = 0;
            }
            else {
                glPolygonMode(GL_FRONT, GL_FILL);
                polygonMode = 1;
            }       
            break; 
        case SDLK_k:
            init_balls();
            break;
            
        default:
            break;
    }
}

void initOpenGL() {

    glShadeModel(GL_SMOOTH);
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glClearDepth(1.0);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glPolygonMode(GL_FRONT, GL_FILL);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    
    glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT1, GL_POSITION, position);
    glLightfv(GL_LIGHT1, GL_SPECULAR, specular);
    glEnable(GL_LIGHT1);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, &light_val);
    glMaterialfv(GL_FRONT, GL_SPECULAR, &light_val);
    glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);
    
}

void resizeWin(int width, int height) {

    glViewport(0, 0, (GLsizei) width, (GLsizei) height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0, ((GLfloat) width / (height == 0 ? 1 : height)), 
                   1.0, 100.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}
void move_balls() {
    
    int i;
    float radius;

    /* Mikachus code :) */

    for (i = 0; i < NR_METABALLS; i++) {
        balls[i].pos.x += balls[i].vel.x;
        radius = (balls[i].radius*2.0);
        if (balls[i].pos.x - radius < 0.0 || balls[i].pos.x + radius > SPACE_SIZE)
            balls[i].vel.x = -balls[i].vel.x;
        else
            balls[i].vel.x += balls[i].acc.x;

        balls[i].pos.y += balls[i].vel.y;
        if (balls[i].pos.y - radius < 0.0 || balls[i].pos.y + radius > SPACE_SIZE)
            balls[i].vel.y = -balls[i].vel.y;
        else
            balls[i].vel.y += balls[i].acc.y;

        
        balls[i].pos.z += balls[i].vel.z;
        if (balls[i].pos.z - radius < 0.0 || balls[i].pos.z + radius > SPACE_SIZE)
            balls[i].vel.z = -balls[i].vel.z;
        else
            balls[i].vel.z += balls[i].acc.z;
            
    }
}

void draw(gridcell_t **cells) {
    
    int i;
    unsigned int sync;
    static double xRotAngle = 0.0, yRotAngle = 0.0, zRotAngle = 0.0;
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glLoadIdentity(); 
    
    glEnable(GL_LIGHTING);
    
    glTranslatef(0.0, 0.0, -60.0);

    
    //glRotatef(xRotAngle, 1.0, 0.0, 0.0);
    //glRotatef(yRotAngle, 0.0, 1.0, 0.0);
    //glRotatef(zRotAngle, 0.0, 0.0, 1.0); 
    glTranslatef(-(SPACE_SIZE/2), -(SPACE_SIZE/2), -(SPACE_SIZE/2));

     
    glBegin(GL_TRIANGLES); 
    for (i = 0; i < NR_METABALLS; i++)    
        make_mesh(cells[i], balls);
    glEnd(); 
    
            
    xRotAngle += 0.5;
    yRotAngle += 0.4;
    zRotAngle += 0.2;
    
    glXGetVideoSyncSGI(&sync);
    glXWaitVideoSyncSGI(2, (sync + 1) % 2, &sync);
   
    SDL_GL_SwapBuffers();

}

int main(int argc, char **argv) {

    SDL_Surface *surface;
    SDL_Event event;
    const SDL_VideoInfo *info;
    int flags;
    char quit = 0;
    gridcell_t **isect;
    

    srand(time(NULL));

       
    init_balls();
    
    init_grid();

    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        fprintf(stderr, "Unable to initialize video: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
    }
   
    atexit(SDL_Quit);
    
    if (!(info = SDL_GetVideoInfo())) {
        fprintf(stderr, "Unable to get video info: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
    }
    
    flags = SDL_OPENGL | SDL_GL_DOUBLEBUFFER | SDL_HWPALETTE | SDL_RESIZABLE;

    if (info->hw_available) {
        flags |= SDL_HWSURFACE;
        printf("Using hardware surface.\n");
    }
    else {
        flags |= SDL_SWSURFACE;
        printf("Using software surface.\n");
    }
    if (info->blit_hw)
        flags |= SDL_HWACCEL;

    
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

    if (!(surface = SDL_SetVideoMode(WIDTH, HEIGHT, BPP, flags))) {
        fprintf(stderr, "Unable to set video mode: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
    }
    
    initOpenGL();
    
    resizeWin(WIDTH, HEIGHT);    
    
    while (!quit) {
        while (SDL_PollEvent(&event)) {
            switch (event.type) {
                case SDL_VIDEORESIZE:
                    if (!(surface = SDL_SetVideoMode(event.resize.w, event.resize.h, BPP, flags))) {
                        fprintf(stderr, "Unable to set video mode: %s\n", SDL_GetError());
                        exit(EXIT_FAILURE);
                        return 1;
                    }
                    resizeWin(event.resize.w, event.resize.h);
                    break;
                case SDL_QUIT:
                    quit = 1;
                    break;
                case SDL_KEYDOWN:
                    keyHandler(&event.key.keysym);
                    break;
            }
        }
       
        isect = find_intersected_cubes(balls);
        draw(isect);
        clean_visited();
        move_balls();
    }
    
    exit(EXIT_SUCCESS);
    
    return 0;

}


